type MaybePromise<T> = T | Promise<T>;
type LazyIterable<T> = MaybePromise<Iterable<T> | AsyncIterable<T>>;

type Fn$c = (...args: any) => any;
declare function compose<T>(fn: Fn$c | LazyIterable<T>, ...fns: (Fn$c | LazyIterable<T>)[]): Fn$c;

type Fn$b = (...args: any) => any;
declare function pipe<T>(...fns: (Fn$b | LazyIterable<T>)[]): Fn$b;

declare function at<T>(index: number): (data: LazyIterable<T>) => unknown;

declare function concat<T>(...data: LazyIterable<T>[]): {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, undefined>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, undefined>;
    [Symbol.asyncIterator]?: undefined;
};

type Fn$a<T> = (input: T, index: number) => boolean;
declare function every<T>(predicate: Fn$a<T>): (data: LazyIterable<T>) => boolean | Promise<boolean>;

type Fn$9<T> = (datum: T, index: number) => boolean;
declare function filter<T>(fn: Fn$9<T>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

type Fn$8<T> = (input: T, index: number) => boolean;
declare function find<T>(predicate: Fn$8<T>): (data: LazyIterable<T>) => (T | undefined) | Promise<T | undefined>;

type Fn$7<T> = (input: T, index: number) => boolean;
declare function findIndex<T>(predicate: Fn$7<T>): (data: LazyIterable<T>) => number | Promise<number>;

type Fn$6<T, R> = (datum: T, index: number) => R;
declare function flatMap<T, R>(fn: Fn$6<T, R>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<unknown, void, undefined>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<unknown, void, undefined>;
    [Symbol.asyncIterator]?: undefined;
} | undefined;

declare function join(separator?: string): (data: LazyIterable<string>) => string | Promise<string> | undefined;

declare function includes<T>(searchElement: T, fromIndex?: number): (data: LazyIterable<T>) => boolean | Promise<boolean>;

type Fn$5<T, R> = (datum: T, index: number) => R;
declare function map<T, R>(fn: Fn$5<T, R>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<R>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<R, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
} | undefined;

type Fn$4<T, R = T> = (acc: R, datum: T, index: number) => R;
declare function reduce<R, T = R>(fn: Fn$4<T, R>, initial: R): (data: LazyIterable<T>) => R | Promise<R> | undefined;

/**
 * This is pretty slow because it has to first go through the whole iterator
 * (to make it an array), then reverse the whole thing and then start
 * yielding again.
 */
declare function reverse<T>(): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<any, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function slice<T>(begin?: number, end?: number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

type Fn$3<T> = (input: T, index: number) => boolean;
declare function some<T>(predicate: Fn$3<T>): (data: LazyIterable<T>) => boolean | Promise<boolean>;

declare function sort<T>(comparator: (a: T, z: T) => number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<any, void, undefined>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function average(): (data: LazyIterable<number>) => any;
declare let mean: typeof average;

declare function batch<T>(size: number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<any, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<any, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
} | undefined;

declare function chunk<T>(size: number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>[], void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T[], void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function compact<T>(): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function delay<T>(ms: number): (data: LazyIterable<T>) => AsyncGenerator<Awaited<T>, void, unknown>;
declare function wait(): (data: LazyIterable<unknown>) => AsyncGenerator<unknown, void, unknown>;

type Options = {
    shallow?: boolean;
};
declare function flatten<T>(options?: Options): (data: LazyIterable<T>) => any;

declare function generate<T>(generator: () => T): Generator<T, void, unknown>;

type KeyFn<T> = (input: T, index: number) => string | number;
declare function groupBy<T>(keySelector: KeyFn<T>): (data: LazyIterable<T>) => Record<string | number, T[]> | Promise<Record<string | number, T[]>>;

declare function head<T>(): (data: LazyIterable<T>) => T | undefined | Promise<T | undefined>;
declare let first: typeof head;

declare function max(): (data: LazyIterable<number>) => number | Promise<number> | undefined;

declare function min(): (data: LazyIterable<number>) => number | Promise<number> | undefined;

type Fn$2<T> = (input: T, index: number) => boolean;
declare function partition<T>(predicate: Fn$2<T>): (data: LazyIterable<T>) => MaybePromise<[T[], T[]]> | undefined;

declare function product(): (data: LazyIterable<number>) => number | Promise<number> | undefined;

declare function range(lowerbound: number, upperbound: number, step?: number): Generator<number, void, unknown>;

declare function skip<T>(amount: number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function sum(): (data: LazyIterable<number>) => number | Promise<number> | undefined;

declare function take<T>(amount: number): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

type Fn$1<T> = (datum: T, index: number) => boolean;
declare function takeWhile<T>(fn: Fn$1<T>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

type Fn<T> = (datum: T, index: number) => void;
declare function tap<T>(fn: Fn<T>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
} | undefined;

declare function toArray<T>(): (data: LazyIterable<T>) => T[] | Promise<T[]> | undefined;

declare function toLength(): (data: LazyIterable<number>) => number | Promise<number> | undefined;

declare function toSet<T>(): (data: LazyIterable<T>) => Set<T> | Promise<Set<T>> | undefined;

declare function unique<T>(): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function where<T>(properties: Record<string | number, any>): (data: LazyIterable<T>) => {
    [Symbol.asyncIterator](): AsyncGenerator<Awaited<T>, void, unknown>;
    [Symbol.iterator]?: undefined;
} | {
    [Symbol.iterator](): Generator<T, void, unknown>;
    [Symbol.asyncIterator]?: undefined;
};

declare function windows<T>(size: number): (data: Iterable<T>) => Generator<T[], void, unknown>;

declare function zip<T>(): (data: Iterable<Iterable<T>>) => Generator<any[], void, unknown>;

export { at, average, batch, chunk, compact, compose, concat, delay, every, filter, find, findIndex, first, flatMap, flatten, generate, groupBy, head, includes, join, map, max, mean, min, partition, pipe, product, range, reduce, reverse, skip, slice, some, sort, sum, take, takeWhile, tap, toArray, toLength, toSet, unique, wait, where, windows, zip };
